#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.

"""
功能: 保存可训练参数功能模块
"""
import os
import time
from collections import OrderedDict

from mindspore import nn
from mindspore import context, Tensor
from mindspore.train.callback._callback import set_cur_net
from mindspore.train.callback import ModelCheckpoint
from mindspore.train.serialization import save_checkpoint, _get_merged_param_data

from mindpet.utils.exceptions import AbsolutePathError, LinkPathError
from mindpet.utils.io_utils import is_link_path
from mindpet.log.log import logger


class TrainableParamsCheckPoint(ModelCheckpoint):
    """TrainableParamsCheckPoint class"""
    def __init__(self, directory, prefix="DELTA_CKP", config=None):
        """
        Callback初始化
        :param directory: ckpt文件的保存路径
        :param prefix: ckpt文件的前缀名称
        :param config: 保存ckpt的配置
        """
        if directory is None or directory == "":
            raise ValueError('Param [directory] is required.')
        if is_link_path(directory):
            raise LinkPathError(f'Path of param [{directory}] is a link path.')
        if not os.path.isabs(directory):
            raise AbsolutePathError(f'Param [{directory}] is not an absolute path.')

        self._last_time_for_keep = time.time()
        self._last_triggered_step = 0
        self._latest_ckpt_file_name = ""
        self._cur_time_for_keep = time.time()
        super().__init__(prefix, directory, config)

    def set_train_info(self, cb_params, cur_step_num):
        """保持ckpt文件始终小于等于配置的ckpt文件最大数"""
        if self._config.keep_checkpoint_max and \
                0 < self._config.keep_checkpoint_max <= self._manager.ckpoint_num:
            self._manager.remove_oldest_ckpoint_file()
        elif self._config.keep_checkpoint_per_n_minutes and self._config.keep_checkpoint_per_n_minutes > 0:
            self._cur_time_for_keep = time.time()
            self._set_per_minute()
        self._last_time_for_keep = time.time()
        self._last_triggered_step = cur_step_num
        if context.get_context("enable_ge"):
            set_cur_net(cb_params.train_network)
            cb_params.train_network.exec_checkpoint_graph()
        # 更新当前epoch数
        if "epoch_num" in self._append_dict:
            self._append_dict["epoch_num"] = self._append_epoch_num + cb_params.cur_epoch_num
        # 更新当前step数
        if "step_num" in self._append_dict:
            self._append_dict["step_num"] = self._append_step_num + cur_step_num

    def trans_network(self, network):
        """从network中选取可训练的参数，仅保存这部分参数"""
        parameter_layout_dict = network.parameter_layout_dict
        network.init_parameters_data()
        param_dict = OrderedDict()
        for param in network.trainable_params():
            param_dict[param.name] = param
        param_list = []
        for param_name, param in param_dict.items():
            each = {"name": param_name}
            param_data = Tensor(param.data.asnumpy())
            if param_name in network.parameter_layout_dict:
                param_data = _get_merged_param_data(network, parameter_layout_dict, param_name, param_data,
                                                    self._config.integrated_save)
            each["data"] = param_data
            param_list.append(each)
        return param_list

    def _set_per_minute(self):
        """只保留每分钟内保存的最后一份ckpt文件"""
        if (self._cur_time_for_keep - self._last_time_for_keep) \
                < self._config.keep_checkpoint_per_n_minutes * 60:
            self._manager.keep_one_ckpoint_per_minutes(self._config.keep_checkpoint_per_n_minutes,
                                                       self._cur_time_for_keep)

    def _save_ckpt(self, cb_params, force_to_save=False):
        """从上下文中获取参数，并且保存ckpt文件"""

        cur_step_num = cb_params.cur_step_num

        if cur_step_num == self._last_triggered_step:
            return

        # 如果参数支持缓存，则在保存ckpt文件之前刷新缓存
        if self._need_flush_from_cache:
            self._flush_from_cache(cb_params)

        enable_save = self._check_save_ckpt(cb_params, force_to_save)

        if not enable_save:
            return

        logger.info("Start to save checkpoint.")

        ckpt_prefix = self._prefix
        if cb_params.batch_num <= 0:
            raise ValueError("[batch_num] can\'t be 0 or less.")

        step_num_in_epoch = int((cur_step_num - 1) % cb_params.batch_num + 1)
        cur_ckpoint_file = ckpt_prefix + "-" + str(cb_params.cur_epoch_num) + "_" \
                           + str(step_num_in_epoch) + ".ckpt"
        ckpt_directory = self._directory
        cur_file = os.path.join(ckpt_directory, cur_ckpoint_file)

        self._manager.update_ckpoint_filelist(ckpt_directory, ckpt_prefix)

        # 训练时维护训练的信息
        self.set_train_info(cb_params, cur_step_num)

        # 获取当前训练的网络
        network = self._config.saved_network if self._config.saved_network is not None else cb_params.train_network

        if isinstance(network, nn.Cell):
            network = self.trans_network(network)
        # 将参数保存为ckpt
        save_checkpoint(network, cur_file, self._config.integrated_save, self._config.async_save,
                        self._append_dict, self._config.enc_key, self._config.enc_mode)
        self._latest_ckpt_file_name = cur_file

        logger.info("Save checkpoint successfully.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Version Control APIs"""
import mindspore as ms
from mindspore import nn
from .version_utils import is_version_ge
if is_version_ge(ms.__version__, '2.0.0'):
    from ..layers.activation import get_activation, _activation
else:
    from mindspore.nn.layer.activation import get_activation, _activation

# pylint: disable=W0127
_activation = _activation
# pylint: disable=W0127
get_activation = get_activation


def get_dropout(dropout_prob):
    if is_version_ge(ms.__version__, '1.11.0'):
        dropout = nn.Dropout(p=dropout_prob)
    else:
        dropout = nn.Dropout(keep_prob=1 - dropout_prob)
    return dropout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) Microsoft Corporation, 2022-2023, All rights reserved.
# Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
# Copyright © Huawei Technologies Co., Ltd. 2010-2023. All rights reserved.
"""R drop Cell"""
import numpy as np
from mindspore import nn
from mindspore.common.tensor import Tensor
from mindspore.common import dtype as mstype
from mindspore import ops


class RDropLoss(nn.Cell):
    """The loss between input and target with r_drop."""
    def __init__(self):
        super().__init__()
        self.cross_entropy_loss = nn.CrossEntropyLoss()
        self.log_softmax = nn.LogSoftmax(axis=-1)
        self.softmax = nn.Softmax(axis=-1)
        self.reshape = ops.Reshape()
        self.gather = ops.Gather()
        self.kl_div = ops.KLDivLoss(reduction='sum')

    def construct(self, logits, label_ids, alpha=4):
        """
        Inputs:
            - **logits** (Tensor) - Predictive value.
            Data type must be mindspore.dtype.float16 or mindspore.dtype.float32.

            - **labels** (Tensor) - Target value.
            Data type must be mindspore.dtype.float16 or mindspore.dtype.float32.

            - **alpha** (int) - Super parameter, specify the disturbance effect.

        Returns:
            Tensor, the computed loss value.
        """
        batch_size = logits.shape[0]
        label_ids = self.reshape(label_ids, (-1,))

        indices_1 = Tensor(np.arange(0, batch_size, 2), mstype.int32)
        indices_2 = Tensor(np.arange(1, batch_size, 2), mstype.int32)

        logits_1 = self.gather(logits, indices_1, 0)
        logits_2 = self.gather(logits, indices_2, 0)

        ce_loss = self.cross_entropy_loss(logits, label_ids)

        kl_loss_1 = self.kl_div(self.log_softmax(logits_1), self.softmax(logits_2))
        kl_loss_2 = self.kl_div(self.log_softmax(logits_2), self.softmax(logits_1))
        kl_loss = (kl_loss_1 + kl_loss_2).mean() / 4 * alpha
        return_value = ce_loss + kl_loss
        return return_value


def rdrop_repeat(*items):
    res = []
    for item in items:
        res.append(item.repeat(2, axis=0))
    return tuple(res)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""quiet option"""

import click
from click.exceptions import BadParameter


class QuietOption(click.core.Option):
    """QuietOption class"""
    def __init__(self):
        super().__init__(
            param_decls=('-q', '--quiet'),
            type=bool,
            is_flag=True,
            help='enable quiet mode, ignore the risk of permission rules of incoming parameters.',
            callback=self.quiet_callback
        )

    # pylint: disable=W0613
    @staticmethod
    def quiet_callback(ctx, params, value):
        # --quiet参数仅允许被定义在首位, 其ctx上下文属性必须为空
        if value is True and ctx.params:
            raise BadParameter('Param [--quiet] should be set first.')

        return value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""ckpt path option"""
import click
from click.exceptions import MissingParameter
from mindpet.utils.constants import PATH_MODE_LIMIT
from mindpet.task.finetune.finetune_options_check import FinetuneOptionsCheckParam, FinetuneOptionsCheck
from mindpet.task.options.path_check_param import default_folder_length_check_param, default_folder_content_check_param


class CkptPathOption(click.core.Option):
    """CkptPathOption class"""
    def __init__(self):
        super().__init__(
            param_decls=('-cp', '--ckpt_path'),
            type=str,
            help='checkpoint local path',
            callback=self.ckpt_path_callback
        )

    @staticmethod
    def ckpt_path_callback(ctx, param, value):
        """
        ckpt_path参数click回调方法

        :param ctx: 上下文信息
        :param param: 参数属性
        :param value: 输入值
        :return: 回调处理后的参数值
        """
        if value is None:
            raise MissingParameter('Param [ckpt_path] is required.')

        check_param = FinetuneOptionsCheckParam(path_length_check_param=default_folder_length_check_param,
                                                path_content_check_param=default_folder_content_check_param,
                                                mode=PATH_MODE_LIMIT,
                                                quiet=ctx.params.get('quiet', False))
        ckpt_path_option = FinetuneOptionsCheck(option_name=param.name, option_value=value)
        ckpt_path_option.check(check_param)

        return ckpt_path_option.option_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""__init__"""

from mindpet.task.options.data_path_option import DataPathOption
from mindpet.task.options.pretrained_model_path_option import PretrainedModelPathOption
from mindpet.task.options.output_path_option import OutputPathOption
from mindpet.task.options.boot_file_path_option import BootFilePathOption
from mindpet.task.options.model_config_path_option import ModelConfigPathOption
from mindpet.task.options.quiet_option import QuietOption
from mindpet.task.options.ckpt_path_option import CkptPathOption
from mindpet.task.options.timeout_option import TimeoutOption



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""activation"""
from __future__ import absolute_import

import numpy as np

from mindspore import _checkparam as validator
from mindspore._extends import cell_attr_register
from mindspore.common import dtype as mstype
from mindspore.common.parameter import Parameter
from mindspore.common.tensor import Tensor
from mindspore.ops import functional as F
from mindspore.ops import operations as P
from mindspore.ops.operations import nn_ops as NN_OPS
from mindspore.nn.cell import Cell
from mindspore import ops
from mindspore.ops.primitive import _primexpr

__all__ = ['Softmin',
           'Softmax',
           'Softmax2d',
           'LogSoftmax',
           'ReLU',
           'ReLU6',
           'RReLU',
           'SeLU',
           'SiLU',
           'Tanh',
           'Tanhshrink',
           'Hardtanh',
           'GELU',
           'FastGelu',
           'Sigmoid',
           'Softsign',
           'PReLU',
           'get_activation',
           'LeakyReLU',
           'HSigmoid',
           'HSwish',
           'ELU',
           'LogSigmoid',
           'LRN',
           'SoftShrink',
           'HShrink',
           'CELU',
           'Threshold',
           'Mish',
           'GLU'
           ]


class CELU(Cell):
    r"""
    Continuously differentiable exponential linear units activation function.

    Applies the continuously differentiable exponential linear units function element-wise.

    .. math::

        \text{CELU}(x) = \max(0,x) + \min(0, \alpha * (\exp(x/\alpha) - 1))

    CELU Activation Function Graph:

    .. image:: images/CELU.png
        :align: center

    Args:
        alpha (float): The :math:`\alpha` value for the Celu formulation. Default: ``1.0`` .

    Inputs:
        - **x** (Tensor) - The input of CELU. The required dtype is float16 or float32.
          The shape is :math:`(N,*)` where :math:`*` means, any number of additional dimensions.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If `alpha` is not a float.
        ValueError: If `alpha` has the value of 0.
        TypeError: If `x` is not a Tensor.
        TypeError: If the dtype of 'input_x' is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-2.0, -1.0, 1.0, 2.0]), mindspore.float32)
        >>> celu = nn.CELU()
        >>> output = celu(x)
        >>> print(output)
        [-0.86466473 -0.63212055  1.          2.        ]
    """

    def __init__(self, alpha=1.0):
        """Initialize CELU."""
        super().__init__()
        self.celu = P.CeLU(alpha=alpha)

    def construct(self, x):
        return self.celu(x)


class Softmin(Cell):
    r"""
    Softmin activation function, which is a two-category function :class:`mindspore.nn.Sigmoid` in the promotion of
    multi-classification, and the purpose is to show the results of multi-classification in the form of probability.

    Calculate the value of the exponential function for the elements of the input Tensor on the `axis`, and then
    normalized to lie in range [0, 1] and sum up to 1.

    Softmin is defined as:

    .. math::
        \text{softmin}(x_{i}) =  \frac{\exp(-x_i)}{\sum_{j=0}^{n-1}\exp(-x_j)},

    where :math:`x_{i}` is the :math:`i`-th slice in the given dimension of the input Tensor.

    Args:
        axis (Union[int, tuple[int]]): The axis to apply Softmin operation, if the dimension of input `x` is x.ndim,
            the range of axis is `[-x.ndim, x.ndim)`. -1 means the last dimension. Default: ``-1`` .

    Inputs:
        - **x** (Tensor) - Tensor for computing Softmin functions with data type of float16 or float32.

    Outputs:
        Tensor, which has the same type and shape as `x` with values in the range [0,1].

    Raises:
        TypeError: If `axis` is neither an int nor a tuple.
        TypeError: If dtype of `x` is neither float16 nor float32.
        ValueError: If `axis` is a tuple whose length is less than 1.
        ValueError: If `axis` is a tuple whose elements are not all in the range [-x.ndim, x.ndim).

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> # axis = -1(default), and the sum of return value is 1.0.
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float16)
        >>> softmin = nn.Softmin()
        >>> output = softmin(x)
        >>> print(output)
        [0.2341  0.636  0.0862  0.01165  0.03168 ]
    """

    def __init__(self, axis=-1):
        """Initialize Softmin."""
        super().__init__()
        self.axis = axis

    def construct(self, x):
        return ops.function.softmin(x, self.axis)


class Softmax2d(Cell):
    r"""
    Softmax function applied to 2D features data.

    Applies `Softmax` to each location :math:`(c, h, w)` with an input Tensor of shape :math:`(C, H, W)` .

    Inputs:
        - **x** (Tensor) - Tensor of shape :math:`(N, C_{in}, H_{in}, W_{in})` or :math:`(C_{in}, H_{in}, W_{in})`.

    Outputs:
        Tensor, which has the same type and shape as `x` with values in the range[0,1].

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.
        ValueError: If `data_format` is neither 'NCHW' nor 'CHW'.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[[[0.1, 0.2]], [[0.3, 0.4]], [[0.6, 0.5]]]]), mindspore.float32)
        >>> softmax2d = nn.Softmax2d()
        >>> output = softmax2d(x)
        >>> print(output)
        [[[[0.25838965 0.28001308]]
        [[0.31559783 0.34200877]]
        [[0.42601252 0.37797815]]]]
    """

    def __init__(self):
        """Initialize Softmax2d."""
        super().__init__()
        self.softmax = P.Softmax(axis=-3)
        self.shape = P.Shape()

    @staticmethod
    @_primexpr
    def _check_input_dim(shape, cls_name):
        dim = len(shape)
        if dim not in (3, 4):
            raise ValueError(f"For '{cls_name}', the in_shape must have 3 or 4 dims, but got {dim}.")

    def construct(self, x):
        x_shape = self.shape(x)
        self._check_input_dim(x_shape, self.cls_name)
        return self.softmax(x)


class Softmax(Cell):
    r"""
    Softmax activation function, which is a two-category function :class:`mindspore.nn.Sigmoid` in the promotion of
    multi-classification, the purpose is to show the results of multi-classification in the form of probability.

    Calculate the value of the exponential function for the elements of the input Tensor on the `axis`, and then
    normalized to lie in range [0, 1] and sum up to 1.

    Softmax is defined as:

    .. math::
        \text{softmax}(x_{i}) =  \frac{\exp(x_i)}{\sum_{j=0}^{n-1}\exp(x_j)},

    where :math:`x_{i}` is the :math:`i`-th slice in the given dimension of the input Tensor.

    Args:
        axis (Union[int, tuple[int]]): The axis to apply Softmax operation, if the dimension of input `x` is x.ndim,
            the range of axis is `[-x.ndim, x.ndim)`, -1 means the last dimension. Default: ``-1`` .

    Inputs:
        - **x** (Tensor) - The input of Softmax with data type of float16 or float32.

    Outputs:
        Tensor, which has the same type and shape as `x` with values in the range[0,1].

    Raises:
        TypeError: If `axis` is neither an int nor a tuple.
        TypeError: If dtype of `x` is neither float16 nor float32.
        ValueError: If `axis` is a tuple whose length is less than 1.
        ValueError: If `axis` is a tuple whose elements are not all in range [-len(x), len(x)).

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> # axis = -1(default), and the sum of return value is 1.0.
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float16)
        >>> softmax = nn.Softmax()
        >>> output = softmax(x)
        >>> print(output)
        [0.03168 0.01166 0.0861  0.636   0.2341 ]
    """

    def __init__(self, axis=-1):
        """Initialize Softmax."""
        super().__init__()
        self.softmax = P.Softmax(axis)

    def construct(self, x):
        return self.softmax(x)


class LogSoftmax(Cell):
    r"""
    Applies the LogSoftmax function to n-dimensional input tensor.

    The input is transformed by the Softmax function and then by the log function to lie in range[-inf,0).

    Logsoftmax is defined as:

    .. math::

        \text{logsoftmax}(x_i) = \log \left(\frac{\exp(x_i)}{\sum_{j=0}^{n-1} \exp(x_j)}\right),

    Args:
        axis (int): The axis to apply LogSoftmax operation, -1 means the last dimension. Default: ``-1`` .

    Inputs:
        - **x** (Tensor) - The input of LogSoftmax, with float16 or float32 data type.

    Outputs:
        Tensor, which has the same type and shape as `x` with output values in the range[-inf,0).

    Raises:
        TypeError: If `axis` is not an int.
        TypeError: If dtype of `x` is neither float16 nor float32.
        ValueError: If `axis` is not in range [-len(x), len(x)).

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[-1.0, 4.0, -8.0], [2.0, -5.0, 9.0]]), mindspore.float32)
        >>> log_softmax = nn.LogSoftmax()
        >>> output = log_softmax(x)
        >>> print(output)
        [[-5.00672150e+00 -6.72150636e-03 -1.20067215e+01]
         [-7.00091219e+00 -1.40009127e+01 -9.12250078e-04]]
    """

    def __init__(self, axis=-1):
        """Initialize LogSoftmax."""
        super().__init__()
        self.log_softmax = P.LogSoftmax(axis)

    def construct(self, x):
        return self.log_softmax(x)


class ELU(Cell):
    r"""
    Exponential Linear Unit activation function.

    Applies the exponential linear unit function element-wise.
    The activation function is defined as:

    .. math::
        E_{i} =
        \begin{cases}
        x_i, &\text{if } x_i \geq 0; \cr
        \alpha * (\exp(x_i) - 1), &\text{otherwise.}
        \end{cases}

    where :math:`x_i` represents the element of the input and :math:`\alpha` represents the `alpha` parameter.

    ELU Activation Function Graph:

    .. image:: images/ELU.png
        :align: center

    Args:
        alpha (float): The alpha value of ELU, the data type is float. Default: ``1.0`` .

    Inputs:
        - **x** (Tensor) - The input of ELU is a Tensor of any dimension with data type of float16 or float32.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If `alpha` is not a float.
        TypeError: If dtype of `x` is neither float16 nor float32.
        ValueError: If `alpha` is not equal to 1.0.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float32)
        >>> elu = nn.ELU()
        >>> result = elu(x)
        >>> print(result)
        [-0.63212055  -0.86466473  0.  2.  1.]
    """

    def __init__(self, alpha=1.0):
        """Initialize ELU."""
        super().__init__()
        self.elu = P.Elu(alpha)

    def construct(self, x):
        return self.elu(x)


class ReLU(Cell):
    r"""
    Rectified Linear Unit activation function.

    .. math::

        \text{ReLU}(x) = (x)^+ = \max(0, x),

    It returns element-wise :math:`\max(0, x)`. Specially, the neurons with the negative output
    will be suppressed and the active neurons will stay the same.

    ReLU Activation Function Graph:

    .. image:: images/ReLU.png
        :align: center

    Inputs:
        - **x** (Tensor) - The input of ReLU is a Tensor of any dimension. The data type is `number <https://www.mind
          spore.cn/docs/en/r2.1/api_python/mindspore.html#mindspore.dtype>`_ .

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If dtype of `x` is not a number.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, 2, -3, 2, -1]), mindspore.float16)
        >>> relu = nn.ReLU()
        >>> output = relu(x)
        >>> print(output)
        [0. 2. 0. 2. 0.]
    """

    def __init__(self):
        """Initialize ReLU."""
        super().__init__()
        self.relu = P.ReLU()

    def construct(self, x):
        return self.relu(x)


class ReLU6(Cell):
    r"""
    Compute ReLU6 activation function.

    ReLU6 is similar to ReLU with a upper limit of 6, which if the inputs are greater than 6, the outputs
    will be suppressed to 6.
    It computes element-wise as

    .. math::

        Y = \min(\max(0, x), 6).

    The input is a Tensor of any valid shape.

    ReLU6 Activation Function Graph:

    .. image:: images/ReLU6.png
        :align: center

    Inputs:
        - **x** (Tensor) - The input of ReLU6 with data type of float16 or float32.

    Outputs:
        Tensor, which has the same type as `x`.

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float16)
        >>> relu6 = nn.ReLU6()
        >>> output = relu6(x)
        >>> print(output)
        [0. 0. 0. 2. 1.]
    """

    def __init__(self):
        """Initialize ReLU6."""
        super().__init__()
        self.relu6 = P.ReLU6()

    def construct(self, x):
        return self.relu6(x)


class LeakyReLU(Cell):
    r"""
    Leaky ReLU activation function.

    The activation function is defined as:

    .. math::
            \text{leaky_relu}(x) = \begin{cases}x, &\text{if } x \geq 0; \cr
            {\alpha} * x, &\text{otherwise.}\end{cases}

    where :math:`\alpha` represents the `alpha` parameter.

    For more details, see `Rectifier Nonlinearities Improve Neural Network Acoustic Models
    <https://ai.stanford.edu/~amaas/papers/relu_hybrid_icml2013_final.pdf>`_.

    LeakyReLU Activation Function Graph:

    .. image:: images/LeakyReLU.png
        :align: center

    Args:
        alpha (Union[int, float]): Slope of the activation function at x < 0. Default: ``0.2`` .

    Inputs:
        - **x** (Tensor) - The input of LeakyReLU is a Tensor of any dimension.

    Outputs:
        Tensor, has the same type and shape as the `x`.

    Raises:
        TypeError: If `alpha` is not a float or an int.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[-1.0, 4.0, -8.0], [2.0, -5.0, 9.0]]), mindspore.float32)
        >>> leaky_relu = nn.LeakyReLU()
        >>> output = leaky_relu(x)
        >>> print(output)
        [[-0.2  4.  -1.6]
         [ 2.  -1.   9. ]]
    """

    def __init__(self, alpha=0.2):
        """Initialize LeakyReLU."""
        super().__init__()
        validator.check_value_type('alpha', alpha, [float, int], self.cls_name)
        self.greater_equal = P.GreaterEqual()
        self.mul = P.Mul()
        self.alpha = alpha
        self.select_op = P.Maximum()
        if self.alpha > 1:
            self.select_op = P.Minimum()

    def construct(self, x):
        alpha_array = P.Cast()(F.scalar_to_tensor(self.alpha), P.DType()(x))
        out = self.select_op(alpha_array * x, x)
        return out


class RReLU(Cell):
    r"""

    Randomized Leaky ReLU activation function.

    The activation function is defined as:

    .. math::
        \text{RReLU}(x_{ji}) = \begin{cases}x_{ji}, &\text{if } x_{ji} \geq 0; \cr
        {\alpha_{ji}} * x_{ji}, &\text{otherwise.}\end{cases}

    where :math:`\alpha_{ji}` ~ :math:`U(l, u)`, :math:`l \le u`.

    Applies the RReLU function elementally, as described in the paper:
    `Empirical Evaluation of Rectified Activations in Convolution Network <https://arxiv.org/pdf/1505.00853.pdf>`_ .

    Args:
        lower (Union[int, float]): Slope of the activation function at x < 0. Default: ``1 / 8`` .
        upper (Union[int, float]): Slope of the activation function at x < 0. Default: ``1 / 3`` .

    Inputs:
        - **x** (Tensor) - The input of RReLU is a Tensor of any dimension.

    Outputs:
        Tensor, after RReLU, has the same type and shape as the `x`.

    Raises:
        TypeError: If `lower` is not a float or an int.
        TypeError: If `upper` is not a float or an int.
        TypeError: If `x` is not a Tensor.
        TypeError: If `x` is not a Tensor of mindspore.float16 or mindpore.float32.
        ValueError: If `lower` is greater than upper.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[-1.0, 4.0], [2.0, 0]]), mindspore.float32)
        >>> r_relu = nn.RReLU()
        >>> output = r_relu(x)
        >>> print(output)
        [[-0.31465699  4.        ]
         [ 2.          0.        ]]
    """

    def __init__(self, lower=1 / 8, upper=1 / 3):
        super().__init__()
        validator.check_value_type('upper', upper, [float, int], self.cls_name)
        validator.check_value_type('lower', lower, [float, int], self.cls_name)
        if lower > upper:
            raise ValueError(f"For {self.cls_name}, the value of 'upper' must be greater than or equal to 'lower', "
                             f"but got upper: {upper}, lower: {lower}. ")
        self.lower = Tensor(lower, dtype=mstype.float32)
        self.upper = Tensor(upper, dtype=mstype.float32)
        self.sign = P.Sign()

    def construct(self, x):
        """RReLU construct."""
        if not isinstance(x, Tensor):
            raise TypeError(f"For 'rrelu', the input must be a Tensor, but got {type(x)}.")
        size = x.shape
        dtype = x.dtype
        sign_matrix = self.sign(x)
        negative_filter = sign_matrix.clip(None, 0)
        positive_filter = sign_matrix.clip(0, None)
        mask = ops.uniform(size, self.lower, self.upper).astype(dtype)
        negative_mask = negative_filter * mask * -1
        total_mask = negative_mask + positive_filter
        out = total_mask * x
        return out


class SeLU(Cell):
    r"""
    Activation function SeLU (Scaled exponential Linear Unit).

    SeLU Activation Function Graph:

    .. image:: images/SeLU.png
        :align: center

    Refer to :func:`mindspore.ops.selu` for more details.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> input_x = Tensor(np.array([[-1.0, 4.0, -8.0], [2.0, -5.0, 9.0]]), mindspore.float32)
        >>> selu = nn.SeLU()
        >>> output = selu(input_x)
        >>> print(output)
        [[-1.1113307 4.202804 -1.7575096]
        [ 2.101402 -1.7462534 9.456309 ]]
    """

    def __init__(self):
        """Initialize SeLU"""
        super().__init__()
        self.selu = P.SeLU()

    def construct(self, input_x):
        return self.selu(input_x)


class SiLU(Cell):
    r"""
    Sigmoid Linear Unit activation function.

    Applies the silu linear unit function element-wise.

    .. math::

        \text{SiLU}(x) = x * \sigma(x),

    where :math:`x_i` is input, :math:`\sigma(x)` is Sigmoid function.

    .. math::

        \text{sigmoid}(x_i) = \frac{1}{1 + \exp(-x_i)},

    SiLU Activation Function Graph:

    .. image:: images/SiLU.png
        :align: center

    Inputs:
        - **x** (Tensor) - Input with the data type float16 or float32.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, 2, -3, 2, -1]), mindspore.float16)
        >>> silu = nn.SiLU()
        >>> output = silu(x)
        >>> print(output)
        [-0.269  1.762  -0.1423  1.762  -0.269]
    """

    def construct(self, x):
        return ops.function.silu(x)


class Tanh(Cell):
    r"""
    Applies the Tanh function element-wise, returns a new tensor with the hyperbolic tangent of the elements of input,
    The input is a Tensor with any valid shape.

    Tanh function is defined as:

    .. math::
        tanh(x_i) = \frac{\exp(x_i) - \exp(-x_i)}{\exp(x_i) + \exp(-x_i)} = \frac{\exp(2x_i) - 1}{\exp(2x_i) + 1},

    where :math:`x_i` is an element of the input Tensor.

    Tanh Activation Function Graph:

    .. image:: images/Tanh.png
        :align: center

    Inputs:
        - **x** (Tensor) - Tensor of any dimension, input with data type of float16 or float32.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([1, 2, 3, 2, 1]), mindspore.float16)
        >>> tanh = nn.Tanh()
        >>> output = tanh(x)
        >>> print(output)
        [0.7617 0.964  0.995  0.964  0.7617]
    """

    def __init__(self):
        """Initialize Tanh."""
        super().__init__()
        self.tanh = P.Tanh()

    def construct(self, x):
        return self.tanh(x)


class Tanhshrink(Cell):
    r"""
    Tanhshrink activation function.

    The tanhshrink function is evaluated by element and returns a new tensor.

    Tanh function is defined as:

    .. math::
        tanhshrink(x_i) =x_i- \frac{\exp(x_i) - \exp(-x_i)}{\exp(x_i) + \exp(-x_i)}
        = x_i-\frac{\exp(2x_i) - 1}{\exp(2x_i) + 1},

    where :math:`x_i` is an element of the input Tensor.

    Inputs:
        - **x** (Tensor) - Tensor of any dimension.

    Outputs:
        Tensor, with the same shape as the `x`.

    Raises:
        TypeError: If `x` is not a Tensor.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([1, 2, 3, 2, 1]), ms.float16)
        >>> tanhshrink = nn.Tanhshrink()
        >>> output = tanhshrink(x)
        >>> print(output)
        [0.2383 1.036  2.004  1.036  0.2383]
    """

    def construct(self, x):
        return F.tanhshrink(x)


class Hardtanh(Cell):
    r"""
    Applies the Hardtanh function element-wise. The activation function is defined as:

    .. math::
        \text{Hardtanh}(x) = \begin{cases}
            1, & \text{ if } x > 1; \\
            -1, & \text{ if } x < -1; \\
            x, & \text{ otherwise. }
        \end{cases}

    Linear region range :math:`[-1, 1]` can be adjusted using `min_val` and `max_val`.

    Hardtanh Activation Function Graph:

    .. image:: images/Hardtanh.png
        :align: center

    Note:
        On Ascend, data type of float16 might lead to accidental accuracy problem.

    Args:
        min_val (Union[int, float]): Minimum value of the linear region range. Default: ``-1.0`` .
        max_val (Union[int, float]): Maximum value of the linear region range. Default: ``1.0`` .

    Inputs:
        - **x** (Tensor) - Input Tensor with data type of float16 or float32.
          On CPU and Ascend support dimension 0-7D. On GPU support dimension 0-4D.

    Outputs:
        Tensor, with the same dtype and shape as `x`.

    Raises:
        TypeError: If `x` is not a Tensor.
        TypeError: If dtype of `x` is neither float16 nor float32.
        TypeError: If dtype of `min_val` is neither float nor int.
        TypeError: If dtype of `max_val` is neither float nor int.
        ValueError: If `min_val` is not less than `max_val`.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float16)
        >>> hardtanh = nn.Hardtanh(min_val=-1.0, max_val=1.0)
        >>> output = hardtanh(x)
        >>> print(output)
        [-1. -1.  0.  1.  1.]
    """

    def __init__(self, min_val=-1.0, max_val=1.0):
        """Initialize Hardtanh."""
        super().__init__()
        self.min_val = min_val
        self.max_val = max_val
        if self.min_val >= self.max_val:
            raise ValueError(f"For Hardtanh, min_val should be less than max_val,"
                             f"but got {self.min_val} and {self.max_val}")

    def construct(self, x):
        return F.hardtanh(x, self.min_val, self.max_val)


class GELU(Cell):
    r"""
    Gaussian error linear unit activation function.

    Applies GELU function to each element of the input. The input is a Tensor with any valid shape.

    GELU is defined as:

    .. math::

        GELU(x_i) = x_i*P(X < x_i),

    where :math:`P` is the cumulative distribution function
    of standard Gaussian distribution and :math:`x_i` is the element of the input.

    GELU Activation Function Graph:

    .. image:: images/GELU.png
        :align: center

    Args:
        approximate (bool): Whether to enable approximation. Default: ``True`` .

            If `approximate` is ``True``, The gaussian error linear activation is:

            :math:`0.5 * x * (1 + tanh(\sqrt(2 / \pi) * (x + 0.044715 * x^3)))`

            else, it is:

            :math:`x * P(X <= x) = 0.5 * x * (1 + erf(x / \sqrt(2)))`, where P(X) ~ N(0, 1).

    Inputs:
        - **x** (Tensor) - The input of GELU with data type of float16 or float32.
          The shape is :math:`(N,*)` where :math:`*` means, any number of additional dimensions.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[-1.0, 4.0, -8.0], [2.0, -5.0, 9.0]]), mindspore.float32)
        >>> gelu = nn.GELU()
        >>> output = gelu(x)
        >>> print(output)
        [[-1.5880802e-01  3.9999299e+00 -3.1077917e-21]
         [ 1.9545976e+00 -2.2918017e-07  9.0000000e+00]]
        >>> gelu = nn.GELU(approximate=False)
        >>> # CPU not support "approximate=False", using "approximate=True" instead
        >>> output = gelu(x)
        >>> print(output)
        [[-1.5865526e-01  3.9998732e+00 -0.0000000e+00]
         [ 1.9544997e+00 -1.4901161e-06  9.0000000e+00]]
    """

    def __init__(self, approximate=True):
        """Initialize GELU."""
        super().__init__()
        validator.check_bool(approximate, 'approximate', self.cls_name)
        self.approximate = approximate
        if self.approximate:
            self.gelu = P.GeLU()
        else:
            self.erf = P.Erf()
            self.sqrt = P.Sqrt()
            self.const0 = Tensor(0.5, mstype.float32)
            self.const1 = Tensor(1.0, mstype.float32)
            self.const2 = Tensor(2.0, mstype.float32)

    def construct(self, x):
        if self.approximate:
            return self.gelu(x)
        return x * F.cast(self.const0, x.dtype) * (F.cast(self.const1, x.dtype) + \
                                                   self.erf(x / self.sqrt(F.cast(self.const2, x.dtype))))


class FastGelu(Cell):
    r"""
    Fast Gaussian error linear unit activation function.

    Applies FastGelu function to each element of the input. The input is a Tensor with any valid shape.

    FastGelu is defined as:

    .. math::
        FastGelu(x_i) = \frac {x_i} {1 + \exp(-1.702 * \left| x_i \right|)} *
                           \exp(0.851 * (x_i - \left| x_i \right|))

    where :math:`x_i` is the element of the input.

    FastGelu Activation Function Graph:

    .. image:: images/FastGelu.png
        :align: center

    Inputs:
        - **x** (Tensor) - The input of FastGelu with data type of float16 or float32.
          The shape is :math:`(N,*)` where :math:`*` means, any number of additional dimensions.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[-1.0, 4.0, -8.0], [2.0, -5.0, 9.0]]), mindspore.float32)
        >>> fast_gelu = nn.FastGelu()
        >>> output = fast_gelu(x)
        >>> print(output)
        [[-1.5418735e-01  3.9921875e+00 -9.7473649e-06]
         [ 1.9375000e+00 -1.0052517e-03  8.9824219e+00]]
    """

    def __init__(self):
        """Initialize FastGelu."""
        super().__init__()
        self.fast_gelu = P.FastGeLU()

    def construct(self, x):
        return self.fast_gelu(x)


class Sigmoid(Cell):
    r"""
    Sigmoid activation function.

    Applies sigmoid-type activation element-wise.

    Sigmoid function is defined as:

    .. math::

        \text{sigmoid}(x_i) = \frac{1}{1 + \exp(-x_i)},

    where :math:`x_i` is the element of the input.

    Sigmoid Activation Function Graph:

    .. image:: images/Sigmoid.png
        :align: center

    Inputs:
        - **input_x** (Tensor) - Tensor of any dimension, the data type is
          float16, float32, float64, complex64 or complex128.

    Outputs:
        Tensor, with the same type and shape as the `input_x`.

    Raises:
        TypeError: If dtype of `input_x` is not float16, float32, float64, complex64 or complex128.
        TypeError: If `input_x` is not a Tensor.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float16)
        >>> sigmoid = nn.Sigmoid()
        >>> output = sigmoid(x)
        >>> print(output)
        [0.2688  0.11914 0.5     0.881   0.7305 ]
    """

    def __init__(self):
        """Initialize Sigmoid."""
        super().__init__()
        self.sigmoid = P.Sigmoid()

    def construct(self, x):
        return self.sigmoid(x)


class Softsign(Cell):
    r"""
    Softsign activation function.

    Softsign Activation Function Graph:

    .. image:: images/Softsign.png
        :align: center

    Refer to :func:`mindspore.ops.softsign` for more details.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([0, -1, 2, 30, -30]), mindspore.float32)
        >>> softsign = nn.Softsign()
        >>> output = softsign(x)
        >>> print(output)
        [ 0.        -0.5         0.6666667  0.9677419 -0.9677419]
    """

    def __init__(self):
        """Initialize Softsign."""
        super().__init__()
        self.softsign = P.Softsign()

    def construct(self, x):
        return self.softsign(x)


class PReLU(Cell):
    r"""
    PReLU activation function.

    Applies the PReLU function element-wise.

    PReLU is defined as:

    .. math::

        PReLU(x_i)= \max(0, x_i) + w * \min(0, x_i),

    where :math:`x_i` is an element of an channel of the input.

    Here :math:`w` is a learnable parameter with a default initial value 0.25.
    Parameter :math:`w` has dimensionality of the argument channel. If called without argument
    channel, a single parameter :math:`w` will be shared across all channels.

    PReLU Activation Function Graph:

    .. image:: images/PReLU.png
        :align: center

    Args:
        channel (int): The elements number of parameter :math:`w`.
          It could be an int, and the value is 1 or the channels number of input tensor `x`. Default: ``1`` .
        w (Union[float, list, Tensor]): The initial value of parameter. It could be a float, a float list or
          a tensor has the same dtype as the input tensor `x`. Default: ``0.25`` .

    Inputs:
        - **x** (Tensor) - The input of PReLU with data type of float16 or float32.
          The shape is :math:`(N, *)` where :math:`*` means, any number of additional dimensions.

    Outputs:
        Tensor, with the same dtype and shape as the `x`.

    Raises:
        TypeError: If `channel` is not an int.
        TypeError: If `w` is not one of a float, a float list, a float Tensor.
        TypeError: If dtype of `x` is neither float16 nor float32.
        ValueError: If the `x` is a 0-D or 1-D Tensor on Ascend.
        ValueError: If `channel` is less than 1.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[[[0.1, 0.6], [0.9, 0.9]]]]), mindspore.float32)
        >>> prelu = nn.PReLU()
        >>> output = prelu(x)
        >>> print(output)
        [[[[0.1 0.6]
           [0.9 0.9]]]]

    """

    @cell_attr_register(attrs="")
    def __init__(self, channel=1, w=0.25):
        """Initialize PReLU."""
        super().__init__()
        validator.check_positive_int(channel, 'channel', self.cls_name)
        if isinstance(w, (float, np.float32)):
            tmp = np.empty((channel,), dtype=np.float32)
            tmp.fill(w)
            w = Tensor(tmp, dtype=mstype.float32)
        elif isinstance(w, list):
            if len(w) != channel:
                raise ValueError(f"For '{self.cls_name}', the length of 'w' must be equal to the 'channel' when "
                                 f"the 'w' is a list, but got the length of 'w': {len(w)}, the 'channel': {channel}.")

            for i in w:
                if not isinstance(i, (float, np.float32)):
                    raise ValueError(f"For '{self.cls_name}', all elements in 'w' must be "
                                     f"float when the 'w' is a list, but got {i}.")
            w = Tensor(w, dtype=mstype.float32)
        elif isinstance(w, Tensor):
            if w.dtype not in (mstype.float16, mstype.float32):
                raise ValueError(f"For '{self.cls_name}', the dtype of 'w' must be float16 or "
                                 f"float32 when the 'w' is a tensor, but got {w.dtype}.")
            if len(w.shape) != 1 or w.shape[0] != channel:
                raise ValueError(f"For '{self.cls_name}', the dimension of 'w' must be 1, and the elements number "
                                 f"should be equal to the 'channel' when the 'w' is a tensor, "
                                 f"but got 'w' shape {w.shape}, the 'channel' {channel}.")
        else:
            raise TypeError(f"For '{self.cls_name}', the 'w' only supported float, list and tensor, "
                            f"but got {type(w).__name__}.")
        self.w = Parameter(w, name='a')
        self.prelu = P.PReLU()

    def construct(self, x):
        return self.prelu(x, F.cast(self.w, x.dtype))


class HSwish(Cell):
    r"""
    Applies hswish-type activation element-wise. The input is a Tensor with any valid shape.

    Hard swish is defined as:

    .. math::
        \text{hswish}(x_{i}) = x_{i} * \frac{ReLU6(x_{i} + 3)}{6},

    HSwish Activation Function Graph:

    .. image:: images/HSwish.png
        :align: center

    Inputs:
        - **x** (Tensor) - The input of HSwish, data type must be float16 or float32.
          The shape is :math:`(N,*)` where :math:`*` means, any number of additional dimensions.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float16)
        >>> hswish = nn.HSwish()
        >>> result = hswish(x)
        >>> print(result)
        [-0.3333 -0.3333  0.      1.667   0.6665]
    """

    def __init__(self):
        """Initialize HSwish."""
        super().__init__()
        self.hswish = P.HSwish()

    def construct(self, x):
        return self.hswish(x)


class HSigmoid(Cell):
    r"""
    Hard sigmoid activation function. Calculates the output according to the input elements.

    Hard sigmoid is defined as:

    .. math::
        \text{hsigmoid}(x_{i}) = \max(0, \min(1, \frac{x_{i} + 3}{6})),

    HSigmoid Activation Function Graph:

    .. image:: images/HSigmoid.png
        :align: center

    Inputs:
        - **input_x** (Tensor) - The input of HSigmoid. Tensor of any dimension.

    Outputs:
        Tensor, with the same type and shape as the `input_x`.

    Raises:
        TypeError: If `input_x` is not a Tensor.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([-1, -2, 0, 2, 1]), mindspore.float16)
        >>> hsigmoid = nn.HSigmoid()
        >>> result = hsigmoid(x)
        >>> print(result)
        [0.3333 0.1666 0.5    0.8335 0.6665]
    """

    def __init__(self):
        """Initialize HSigmoid."""
        super().__init__()
        self.hsigmoid = P.HSigmoid()

    def construct(self, input_x):
        return self.hsigmoid(input_x)


class LogSigmoid(Cell):
    r"""
    Applies logsigmoid activation element-wise. The input is a Tensor with any valid shape.

    Logsigmoid is defined as:

    .. math::
        \text{logsigmoid}(x_{i}) = \log(\frac{1}{1 + \exp(-x_i)}),

    where :math:`x_{i}` is the element of the input.

    LogSigmoid Activation Function Graph:

    .. image:: images/LogSigmoid.png
        :align: center

    Inputs:
        - **x** (Tensor) - The input of LogSigmoid with data type of float16 or float32.
          The shape is :math:`(N,*)` where :math:`*` means, any number of additional dimensions.

    Outputs:
        Tensor, with the same type and shape as the `x`.

    Raises:
        TypeError: If dtype of `x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> net = nn.LogSigmoid()
        >>> x = Tensor(np.array([1.0, 2.0, 3.0]), mindspore.float32)
        >>> output = net(x)
        >>> print(output)
        [-0.31326166 -0.12692806 -0.04858734]
    """

    def __init__(self):
        """Initialize LogSigmoid."""
        super().__init__()
        self.mul = P.Mul()
        self.exp = P.Exp()
        self.add = P.Add()
        self.rec = P.Reciprocal()
        self.log = P.Log()

    def construct(self, input_x):
        neg_input = self.mul(input_x, -1)
        exp_neg_input = self.exp(neg_input)
        exp_neg_input_1 = self.add(exp_neg_input, 1)
        rec_exp_neg_input_1 = self.rec(exp_neg_input_1)
        ret = self.log(rec_exp_neg_input_1)
        return ret


class LRN(Cell):
    r"""
    Local Response Normalization.

    Refer to :func:`mindspore.ops.lrn` for more details.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> input_x = Tensor(np.array([[[[0.1], [0.2]],
        ...                       [[0.3], [0.4]]]]), mindspore.float32)
        >>> output = nn.LRN()(input_x)
        >>> print(output)
        [[[[0.09534626]
           [0.1825742 ]]
          [[0.2860388 ]
           [0.3651484 ]]]]
    """

    def __init__(self, depth_radius=5, bias=1.0, alpha=1.0, beta=0.5, norm_region="ACROSS_CHANNELS"):
        """Initialize LRN."""
        super().__init__()
        self.lrn_op = NN_OPS.LRN(depth_radius, bias, alpha, beta, norm_region)

    def construct(self, input_x):
        return self.lrn_op(input_x)


class SoftShrink(Cell):
    r"""
    Applies the SoftShrink function element-wise.

    .. math::
        \text{SoftShrink}(x) =
        \begin{cases}
        x - \lambda, & \text{ if } x > \lambda \\
        x + \lambda, & \text{ if } x < -\lambda \\
        0, & \text{ otherwise }
        \end{cases}

    SoftShrink Activation Function Graph:

    .. image:: images/Softshrink.png
        :align: center

    Args:
        lambd (float): the :math:`\lambda` must be no less than zero for the SoftShrink formulation.
            Default: ``0.5`` .

    Inputs:
        - **input_x** (Tensor) - The input of SoftShrink with data type of float16 or float32.
          Any number of additional dimensions.

    Outputs:
        Tensor, has the same shape and data type as `input_x`.

    Raises:
        TypeError: If lambd is not a float.
        TypeError: If input_x is not a Tensor.
        TypeError: If dtype of input_x is neither float16 nor float32.
        ValueError: If lambd is less than 0.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> input_x = Tensor(np.array([[ 0.5297,  0.7871,  1.1754], [ 0.7836,  0.6218, -1.1542]]), mindspore.float16)
        >>> softshrink = nn.SoftShrink()
        >>> output = softshrink(input_x)
        >>> print(output)
        [[ 0.02979  0.287    0.676  ]
         [ 0.2837   0.1216  -0.6543 ]]
    """

    def __init__(self, lambd=0.5):
        super().__init__()
        self.softshrink = P.SoftShrink(lambd)

    def construct(self, input_x):
        output = self.softshrink(input_x)
        return output


class HShrink(Cell):
    r"""
    Hard Shrink activation function. Calculates the output according to the input elements.

    The formula is defined as follows:

    .. math::
        \text{HardShrink}(x) =
        \begin{cases}
        x, & \text{ if } x > \lambda \\
        x, & \text{ if } x < -\lambda \\
        0, & \text{ otherwise }
        \end{cases}

    HShrink Activation Function Graph:

    .. image:: images/HShrink.png
        :align: center

    Args:
        lambd (float): The threshold :math:`\lambda` defined by the Hard Shrink formula. Default: ``0.5`` .

    Inputs:
        - **input_x** (Tensor) - The input of Hard Shrink with data type of float16 or float32.

    Outputs:
        Tensor, the same shape and data type as the input.

    Raises:
        TypeError: If `lambd` is not a float.
        TypeError: If dtype of `input_x` is neither float16 nor float32.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> input_x = Tensor(np.array([[ 0.5,  1,  2.0], [0.0533,0.0776,-2.1233]]), mindspore.float32)
        >>> hshrink = nn.HShrink()
        >>> output = hshrink(input_x)
        >>> print(output)
        [[ 0.      1.      2.    ]
        [ 0.      0.     -2.1233]]
    """

    def __init__(self, lambd=0.5):
        super().__init__()
        self.hshrink = P.HShrink(lambd)

    def construct(self, input_x):
        return self.hshrink(input_x)


class Threshold(Cell):
    r"""Thresholds each element of the input Tensor.

    The formula is defined as follows:

    .. math::
        y =
        \begin{cases}
        x, &\text{ if } x > \text{threshold} \\
        \text{value}, &\text{ otherwise }
        \end{cases}

    Args:
        threshold (Union[int, float]): The value to threshold at.
        value (Union[int, float]): The value to replace with when element is less than threshold.

    Inputs:
        - **input_x** (Tensor) - The input of Threshold with data type of float16 or float32.

    Outputs:
        Tensor, the same shape and data type as the input.

    Raises:
        TypeError: If `threshold` is not a float or an int.
        TypeError: If `value` is not a float or an int.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> m = nn.Threshold(0.1, 20)
        >>> inputs = Tensor([0.1, 0.2, 0.3], mindspore.float32)
        >>> outputs = m(inputs)
        >>> print(outputs)
        [ 20.0     0.2      0.3]
    """

    def __init__(self, threshold, value):
        """Initialize Threshold."""
        super().__init__()
        self.threshold = threshold
        self.value = value

    def construct(self, input_x):
        return F.threshold(input_x, self.threshold, self.value)


class Mish(Cell):
    r"""
    Computes MISH(A Self Regularized Non-Monotonic Neural Activation Function) of input tensors element-wise.

    Refer to :func:`mindspore.ops.mish` for more details.

    Mish Activation Function Graph:

    .. image:: images/Mish.png
        :align: center

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore
        >>> from mindspore import Tensor, nn
        >>> import numpy as np
        >>> x = Tensor(np.array([[-1.0, 4.0, -8.0], [2.0, -5.0, 9.0]]), mindspore.float32)
        >>> mish = nn.Mish()
        >>> output = mish(x)
        >>> print(output)
        [[-3.03401530e-01 3.99741292e+00 -2.68321624e-03]
         [ 1.94395900e+00 -3.35762873e-02 9.00000000e+00]]
    """

    def __init__(self):
        """Initialize Mish."""
        super().__init__("Mish")
        self.mish = NN_OPS.Mish()

    def construct(self, input_x):
        return self.mish(input_x)


class GLU(Cell):
    r"""
    The gated linear unit function.

    .. math::
        {GLU}(a, b)= a \otimes \sigma(b)

    where :math:`a` is the first half of the input matrices and :math:`b` is the second half.

    Here :math:`\sigma` is the sigmoid function, and :math:`\otimes` is the Hadamard product.

    Args:
        axis (int): the axis to split the input. Default: ``-1`` , the last axis in `x`.

    Inputs:
        - **x** (Tensor) - :math:`(\ast_1, N, \ast_2)` where `*` means, any number of additional dimensions.

    Outputs:
        Tensor, the same dtype as the `x`, with the shape :math:`(\ast_1, M, \ast_2)` where :math:`M=N/2`.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore as ms
        >>> m = ms.nn.GLU()
        >>> input = ms.Tensor([[0.1,0.2,0.3,0.4],[0.5,0.6,0.7,0.8]])
        >>> output = m(input)
        >>> print(output)
        [[0.05744425 0.11973753]
         [0.33409387 0.41398472]]
    """

    def __init__(self, axis=-1):
        """Initialize GLU."""
        super().__init__("GLU")
        self.dim = axis
        self.spilt = P.Split(axis=axis, output_num=2)
        self.sigmoid = P.Sigmoid()

    def construct(self, x):
        x1, x2 = self.spilt(x)
        x2 = self.sigmoid(x2)
        return x1 * x2


_activation = {
    'softmin': Softmin,
    'softmax': Softmax,
    'softmax2d': Softmax2d,
    'logsoftmax': LogSoftmax,
    'relu': ReLU,
    'relu6': ReLU6,
    'rrelu': RReLU,
    'silu': SiLU,
    'tanh': Tanh,
    'tanhshrink': Tanhshrink,
    'hardtanh': Hardtanh,
    'gelu': GELU,
    'fast_gelu': FastGelu,
    'elu': ELU,
    'sigmoid': Sigmoid,
    'softsign': Softsign,
    'prelu': PReLU,
    'leakyrelu': LeakyReLU,
    'hswish': HSwish,
    'hsigmoid': HSigmoid,
    'logsigmoid': LogSigmoid,
    'softshrink': SoftShrink,
    'hshrink': HShrink,
    'threshold': Threshold,
    'mish': Mish,
}


def get_activation(name, prim_name=None):
    """
    Gets the activation function.

    Args:
        name (str): The name of the activation function.
        prim_name (Union[str, None]): The name of primitive. Default: ``None`` .

    Returns:
        Function, the activation function.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Examples:
        >>> import mindspore.nn as nn
        >>> sigmoid = nn.get_activation('sigmoid')
        >>> print(sigmoid)
        Sigmoid<>
    """
    msg_prefix = f"For '{prim_name}', the" if prim_name else "The"
    if name is None:
        return None

    if name not in _activation:
        raise KeyError(f"{msg_prefix} 'name' must be in {list(_activation.keys())}, but got {name}.")
    return _activation[name]()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2022-2023, All rights reserved.
"""LoRA Cell"""
import math
from mindspore import nn
from mindspore import ops
from mindspore import Parameter
from mindspore.common.tensor import Tensor
from mindspore import dtype as mstype
from mindspore.ops import operations as P
from mindspore.ops import functional as F
try:
    from mindspore._checkparam import Validator, Rel
    INC_LEFT = Rel.INC_LEFT
except ImportError:
    import mindspore._checkparam as Validator
    INC_LEFT = Validator.INC_LEFT
from mindspore.common.initializer import initializer, HeUniform
from mindpet.delta.delta_constants import VALID_TENSOR_DATATYPE
from mindpet.utils.version_control import get_dropout


class LoRADense(nn.Dense):
    """Define a dense layer with LoRA structure.

    Attributes:
        lora_in_channels (int): The number of channels in the input space.
        lora_out_channels (int): The number of channels in the output space.
        lora_rank(int): The number of rows(columns) in LoRA matrices.
        lora_alpha(float): A constant in lora_rank.
        lora_dropout(float): The dropout rate, greater equal than 0 and less than 1.
        param_init_type(:class:`mindspore.dtype`): The type of data in initialized tensor.
        compute_dtype(:class:`mindspore.dtype`): The compute type of data.

    """

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            lora_rank: int,
            lora_alpha: float,
            lora_dropout: float = 0.,
            lora_a_init=HeUniform(negative_slope=math.sqrt(5)),
            lora_b_init='zeros',
            param_init_type=mstype.float32,
            compute_dtype=mstype.float16,
            **kwargs
    ):
        super().__init__(in_channels, out_channels, **kwargs)

        # Check params
        self._check_num(lora_rank, lora_alpha, lora_dropout)
        self._check_init(lora_a_init, lora_b_init, lora_rank)
        self._check_type_of_data(param_init_type, compute_dtype)

        # Define and initialize params
        self.lora_rank = lora_rank
        self.lora_alpha = lora_alpha
        self.lora_dropout = get_dropout(lora_dropout)
        self.mindpet_delta_lora_a = Parameter(
            initializer(lora_a_init, [lora_rank, in_channels], param_init_type),
            name='mindpet_delta_lora_A')
        self.mindpet_delta_lora_b = Parameter(initializer(lora_b_init, [out_channels, lora_rank], param_init_type),
                                         name='mindpet_delta_lora_B')
        self.scaling = self.lora_alpha / self.lora_rank

        # Calculation utils
        self.mul = ops.Mul()
        self.add = ops.Add()
        self.cast = ops.Cast()
        self.dtype = compute_dtype
        self.lora_a_matmul = P.MatMul(transpose_b=True)
        self.lora_b_matmul = P.MatMul(transpose_b=True)

    def construct(self, input_tensor):
        """Foward"""
        # Data type operation
        ori_dtype = F.dtype(input_tensor)
        input_tensor = self.cast(input_tensor, self.dtype)
        weight = self.cast(self.weight, self.dtype)
        lora_a = self.cast(self.mindpet_delta_lora_a, self.dtype)
        lora_b = self.cast(self.mindpet_delta_lora_b, self.dtype)
        scaling = self.cast(self.scaling, self.dtype)

        # Shape operations
        x_shape = self.shape_op(input_tensor)
        nn.layer.basic.check_dense_input_shape(x_shape, self.cls_name)
        input_tensor = self.reshape(input_tensor, (-1, x_shape[-1]))

        # Dense result
        dense_result = self.matmul(input_tensor, weight)
        if self.has_bias:
            bias = self.cast(self.bias, self.dtype)
            dense_result = self.bias_add(dense_result, bias)

        # LoRA result
        input_tensor = self.lora_dropout(input_tensor)
        input_tensor = self.lora_a_matmul(input_tensor, lora_a)
        input_tensor = self.lora_b_matmul(input_tensor, lora_b)
        input_tensor = self.mul(input_tensor, scaling)

        # Result addition and activation
        dense_result = self.add(dense_result, input_tensor)
        if self.activation_flag:
            dense_result = self.activation(dense_result)

        # Shape restore
        if len(x_shape) != 2:
            out_shape = x_shape[:-1] + (-1,)
            dense_result = self.reshape(dense_result, out_shape)
        dense_result = self.cast(dense_result, ori_dtype)

        return dense_result

    def shard(self, strategy_org_dense_matmul=None,
              strategy_org_bias_add=None,
              strategy_lora_dropout=None,
              strategy_lora_a_matmul=None,
              strategy_lora_b_matmul=None,
              strategy_lora_mul=None,
              strategy_lora_add=None,
              strategy_activation=None):
        r"""
        Set the shard for the LoraDense. All the strategy size should be equal to the inputs.

        Note:
            It is valid only in semi auto parallel or auto parallel mode.
            In other parallel modes, strategies set here will be ignored.

        Args:
            strategy_org_dense_matmul (tuple): The strategy for the origin dense matmul.
            strategy_org_bias_add (tuple): The strategy for the origin dense bias_add.
            strategy_lora_dropout (tuple): The strategy for the LoraDense dropout.
            strategy_lora_a_matmul (tuple): The strategy for the LoraDense lora_a matmul.
            strategy_lora_b_matmul (tuple): The strategy for the LoraDense lora_b matmul.
            strategy_lora_mul (tuple): The strategy for the LoraDense mul.
            strategy_lora_add (tuple): The strategy for the LoraDense add.
            strategy_activation (tuple): The strategy for the strategy_activation.
        """
        try:
            self.matmul.shard(strategy_org_dense_matmul)
            if self.has_bias:
                self.bias_add.shard(strategy_org_bias_add)
            self.lora_dropout.dropout.shard(strategy_lora_dropout)
            self.lora_a_matmul.shard(strategy_lora_a_matmul)
            self.lora_b_matmul.shard(strategy_lora_b_matmul)
            self.mul.shard(strategy_lora_mul)
            self.add.shard(strategy_lora_add)

            if self.activation_flag:
                if isinstance(self.activation, nn.LeakyReLU):
                    self.activation.select_op.shard((strategy_activation[0], strategy_activation[0]))
                elif isinstance(self.activation, nn.LogSigmoid):
                    self.activation.mul.shard((strategy_activation[0], ()))
                    self.activation.exp.shard(strategy_activation)
                    self.activation.add.shard((strategy_activation[0], ()))
                    self.activation.rec.shard(strategy_activation)
                    self.activation.log.shard(strategy_activation)
                elif isinstance(self.activation, nn.LogSoftmax):
                    raise ValueError("The 'LogSoftmax' function is not supported in semi auto parallel "
                                     "or auto parallel mode.")
                else:
                    getattr(self.activation, self.act_name).shard(strategy_activation)
        except Exception as ex:
            # pylint: disable=W0719
            raise Exception(f"Exception occurred when set the shard for LoRADense, error message: {str(ex)}") from ex

    def _check_num(self, lora_rank, lora_alpha, lora_dropout):
        Validator.check_positive_int(arg_value=lora_rank, arg_name="lora_rank", prim_name=self.cls_name)

        Validator.check_value_type(arg_value=lora_alpha, arg_name='lora_alpha',
                                   valid_types=[int, float], prim_name=self.cls_name)
        if lora_alpha == 0:
            raise ValueError(f"For {self.cls_name}, the 'lora_alpha' cannot be zero, "
                             f"but got '{lora_alpha}' with type {type(lora_alpha)}.")

        Validator.check_float_range(arg_value=lora_dropout, lower_limit=0.0, upper_limit=1.0,
                                    rel=INC_LEFT, arg_name='lora_dropout', prim_name=self.cls_name)

    def _check_init(self, lora_a_init, lora_b_init, lora_rank):
        """check init arguments"""
        if isinstance(lora_a_init, Tensor):
            if lora_a_init.ndim != 2 or lora_a_init.shape[0] != lora_rank or \
                    lora_a_init.shape[1] != self.in_channels:
                raise ValueError(f"For '{self.cls_name}', [lora_a_init] shape error. "
                                 f"The ndim of [lora_a_init] must be equal to 2, "
                                 f"the first dim must be equal to 'lora_rank', "
                                 f"and the second dim must be equal to 'in_channels'. ")
        if isinstance(lora_b_init, Tensor):
            if lora_b_init.ndim != 2 or lora_b_init.shape[0] != self.out_channels or \
                    lora_b_init.shape[1] != lora_rank:
                raise ValueError(f"For '{self.cls_name}', [lora_b_init] shape error. "
                                 f"The ndim of [lora_b_init] must be equal to 2, "
                                 f"the first dim must be equal to 'out_channels', "
                                 f"and the second dim must be equal to 'lora_rank'. ")

    def _check_type_of_data(self, param_init_type, compute_dtype):
        if param_init_type not in VALID_TENSOR_DATATYPE:
            raise TypeError(f"For {self.cls_name}, the 'param_init_type' must be mindspore.dtype.float16 or "
                            f"mindspore.dtype.float32, but got {param_init_type}.")
        if compute_dtype not in VALID_TENSOR_DATATYPE:
            raise TypeError(f"For {self.cls_name}, the 'compute_dtype' must be mindspore.dtype.float16 or "
                            f"mindspore.dtype.float32, but got {compute_dtype}.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""log module."""

import logging
import logging.config
import logging.handlers
import os
import sys
from typing import Dict, List, Tuple, Union
import traceback

try:
    from concurrent_log_handler import ConcurrentRotatingFileHandler as RFHandler
except ImportError:
    from logging.handlers import RotatingFileHandler as RFHandler

from mindpet.log.log_utils import check_list, const, convert_nodes_devices_input, create_dirs, \
    generate_rank_list, get_num_nodes_devices, get_rank_info, log_args_black_list_characters_replace, check_link_path
from mindpet.utils.constants import DEFAULT_MAX_LOG_FILE_NUM, DEFAULT_MAX_LOG_FILE_SIZE, ABNORMAL_EXIT_CODE
from mindpet.utils.exceptions import MakeDirError, UnsupportedPlatformError, PathOwnerError, PathModeError


logger_list = {}
stream_handler_list = {}
file_handler_list = {}
LOG_RECORD_MAX_LEN = 2048
LEVEL_CONVERTER = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL
}
LOGGER_SOURCE_LIST = ('SDK', 'CLI')
UNKNOWN = 'unknown'
MODEL_LOG_FORMATTER = logging.Formatter(const.default_model_log_format)
MODE_640 = 0o640
MODE_440 = 0o440
MODE_750 = 0o750


def judge_stdout(rank_id: int,
                 rank_size: int,
                 is_output: bool,
                 nodes: Union[List, Tuple, Dict[str, int], None] = None,
                 devices: Union[List, Tuple, Dict[str, int], None] = None) -> bool:
    """判断日志内容是否需要输出到屏幕.

    Args:
        rank_id (int): 节点Rank id.
        rank_size (int): Rank size.
        is_output (bool): 是否需要输出到屏幕
        nodes (list or tuple or dict or None): 节点列表，列表中的节点的日志会输出到屏幕
        devices (list or tuple or dict or None): 设备列表，列表中的设备的日志会输出到屏幕

    Returns:
        is_output (bool): 最终判定是否需要输出到屏幕
    """
    nodes_or_devices = (nodes is not None or devices is not None)
    if is_output and rank_size > 1 and nodes_or_devices:
        num_nodes, num_devices = get_num_nodes_devices(rank_size)
        stdout_nodes = convert_nodes_devices_input(nodes, num_nodes)
        stdout_devices = convert_nodes_devices_input(devices, num_devices)
        check_list('nodes', stdout_nodes, num_nodes)
        check_list('devices', stdout_devices, num_devices)
        rank_list = generate_rank_list(stdout_nodes, stdout_devices)
        if rank_id not in rank_list:
            is_output = False

    return is_output


def validate_nodes_devices_input(var_name: str, var):
    """检查节点或者设备的参数输入

    Args:
        var_name (str): 参数名，节点或设备
        var: 参数

    Returns:
        None
    """
    if not (var is None or isinstance(var, (list, tuple, dict))):
        raise TypeError(f'The value of {var_name} can be None or a value of type tuple, list, or dict.')
    if isinstance(var, (list, tuple)):
        for item in var:
            if not isinstance(item, int):
                raise TypeError('The elements of a variable of type list or tuple must be of type int.')


def validate_level(var_name: str, var):
    """验证日志的等级是否正确

    Args:
        var_name (str): 参数名称
        var: 参数

    Returns:
        None
    """
    if not isinstance(var, str):
        raise TypeError(f'The format of {var_name} must be of type str.')
    if var not in const.level:
        raise ValueError(f'{var_name}={var} needs to be in {const.level}')


def validate_std_input_format(to_std: bool, stdout_nodes: Union[List, Tuple, None],
                              stdout_devices: Union[List, Tuple, None], stdout_level: str):
    """验证日志内容输出的节点、设备、等级"""

    if not isinstance(to_std, bool):
        raise TypeError('The format of the to_std must be of type bool.')

    validate_nodes_devices_input('stdout_nodes', stdout_nodes)
    validate_nodes_devices_input('stdout_devices', stdout_devices)
    validate_level('stdout_level', stdout_level)


def validate_file_input_format(file_level: Union[List, Tuple], file_save_dir: str, append_rank_dir: bool,
                               file_name: Union[List, Tuple]):
    """验证日志内容落盘的参数"""

    if not isinstance(file_level, (tuple, list)):
        raise TypeError('The value of file_level should be list or a tuple.')
    for level in file_level:
        validate_level('level in file_level', level)

    if not isinstance(file_name, (tuple, list)):
        raise TypeError('The value of file_name should be a list or a tuple.')

    if not len(file_level) == len(file_name):
        raise ValueError('The length of file_level and file_name should be equal.')

    if not isinstance(file_save_dir, str):
        raise TypeError('The type of file_save_dir should be a value of type str.')

    if not isinstance(append_rank_dir, bool):
        raise TypeError('The value of append_rank_dir should be a value of type bool.')

    for name in file_name:
        if not isinstance(name, str):
            raise TypeError('The value of name in file_name should be a value of type str.')
        if name.startswith('/'):
            raise ValueError('The file name cannot start with "/".')
        if name.startswith('../'):
            raise ValueError('The file name cannot start with "../".')


def _convert_level(level: str) -> int:
    """转换日志等级参数

    Args:
        level (str): 用户定义的日志等级

    Returns:
        level (str): 转换后的日志等级
    """
    logging_level = LEVEL_CONVERTER.get(level, logging.INFO)

    return logging_level


def get_stream_handler(stdout_format: str, stdout_level: str):
    """获得日志的流处理器"""
    if not stdout_format:
        stdout_format = const.default_stdout_format
    handler_name = f'{stdout_format}.{stdout_level}'
    if handler_name in stream_handler_list:
        return stream_handler_list.get(handler_name)

    stream_handler = CustomedStreamHandler(sys.stdout)
    stream_handler.setLevel(_convert_level(stdout_level))
    stream_formatter = logging.Formatter(stdout_format)
    stream_handler.setFormatter(stream_formatter)

    stream_handler_list[handler_name] = stream_handler

    return stream_handler


class CustomedStreamHandler(logging.StreamHandler):
    def format(self, record):
        return log_format(self, record)


def log_format(handler, record):
    if record.__dict__.get("Model"):
        fmt = MODEL_LOG_FORMATTER
    else:
        fmt = handler.formatter
    return fmt.format(record)


def get_file_path_list(base_save_dir: str,
                       append_rank_dir: bool,
                       server_id: int,
                       rank_id: int,
                       file_name: Union[Tuple, List],
                       mode: int = MODE_750) -> List:
    """获得日志落盘的日志文件路径"""
    if not base_save_dir:
        base_save_dir = os.path.expanduser(const.local_default_log_file_dir)

    file_save_dir = base_save_dir
    if append_rank_dir:
        rank_str = const.rank_dir_format.format(rank_id)
        node_str = const.node_dir_format.format(server_id)
        temp_save_dir = os.path.join(base_save_dir, node_str)
        file_save_dir = os.path.join(temp_save_dir, rank_str)

    file_path = []
    for name in file_name:
        path = os.path.join(file_save_dir, name)
        path = os.path.realpath(path)
        base_dir = os.path.dirname(path)
        create_dirs(base_dir)
        file_path.append(path)

    for root, dirs, _ in os.walk(base_save_dir):
        for dir_name in dirs:
            path = os.path.join(root, dir_name)
            if not check_link_path(path):
                os.chmod(path, mode)

    return file_path


class CustomizedRotatingFileHandler(RFHandler):
    """继承concurrent_log_handler包下的ConcurrentRotatingFileHandler类
    1.实现控制日志文件的并发写入和读取
    2.实现日志文件的权限控制
    3.日志文件的绕接
    """
    def doRollover(self) -> None:
        """重写doRoller，实现绕接，并且对日志文件的权限进行控制"""
        super().doRollover()
        if self.backupCount > 0:
            for i in range(self.backupCount - 1, 0, -1):
                file_name = self.rotation_filename(f'{self.baseFilename}.{i}')
                if os.path.exists(file_name) and not check_link_path(str(file_name)):
                    os.chmod(file_name, MODE_440)

    def emit(self, record) -> None:
        """重写emit方法，实现日志内容长度限制"""
        tmp_out = record.getMessage()
        if tmp_out is not None and len(tmp_out) > LOG_RECORD_MAX_LEN:
            record.msg = tmp_out[:LOG_RECORD_MAX_LEN]
            record.args = ()
        super().emit(record)
        if not check_link_path(str(self.baseFilename)) and os.path.exists(self.baseFilename):
            os.chmod(self.baseFilename, MODE_640)

    def format(self, record):
        return log_format(self, record)

    def _open_lockfile(self):
        super()._open_lockfile()
        self._do_chmod(MODE_640)

    def _do_chmod(self, mode):
        if mode is not None and os.path.exists(self.lockFilename):
            os.chmod(self.lockFilename, mode)


def get_file_handler_list(file_level: Union[List, Tuple], file_path: Union[List, Tuple], max_file_size: int,
                          max_num_of_files: int) -> List:
    """获得日志的文件处理器"""
    logging_level = []
    for level in file_level:
        logging_level.append(_convert_level(level))

    max_file_size = max_file_size * 1024 * 1024

    file_formatter = logging.Formatter(const.default_filehandler_format)

    file_handlers = []
    for path, level in zip(file_path, logging_level):
        handler_name = f'{path}.{max_file_size}.{max_num_of_files}.{level}'

        if handler_name in file_handler_list:
            file_handlers.append(file_handler_list.get(handler_name))
        else:
            file_handler = CustomizedRotatingFileHandler(filename=path,
                                                         maxBytes=max_file_size,
                                                         backupCount=max_num_of_files,
                                                         delay=True)
            file_handler.setLevel(level)
            file_handler.setFormatter(file_formatter)
            file_handlers.append(file_handler)

            file_handler_list[handler_name] = file_handler
    return file_handlers


class MxLogger(logging.Logger):
    """Define Mx logger"""
    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.source = None
        self.method = None
        self.propagate = False
        self.to_std = kwargs.get('to_std', True)
        self.stdout_nodes = kwargs.get('stdout_nodes', None)
        self.stdout_devices = kwargs.get('stdout_devices', (0,))
        self.stdout_level = kwargs.get('stdout_level', 'INFO')
        self.stdout_format = kwargs.get('stdout_format', '')
        self.file_level = kwargs.get('file_level', ('INFO', 'ERROR'))
        self.append_rank_dir = kwargs.get('append_rank_dir', True)
        self.file_name = kwargs.get('file_name', ('mx_logger.log', 'error.log'))
        self.max_file_size = kwargs.get('max_file_size', DEFAULT_MAX_LOG_FILE_SIZE)
        self.max_num_of_files = kwargs.get('max_num_of_files', DEFAULT_MAX_LOG_FILE_NUM)
        self.set_flag = False

    def makeRecord(
            self,
            name,
            level,
            fn,
            lno,
            msg,
            args,
            exc_info,
            func=None,
            extra=None,
            sinfo=None,
    ):
        """重写makeRecord方法，日志内容中增加发起端标识和pid信息"""
        if extra is None:
            extra = {}
        extra['source'] = self.source if hasattr(self, 'source') else UNKNOWN
        extra['pid'] = self.pid if hasattr(self, 'pid') else UNKNOWN
        args = log_args_black_list_characters_replace(args)
        return super().makeRecord(name, level=level, fn=fn, lno=lno, msg=msg, args=args,
                                                exc_info=exc_info,
                                                func=func, extra=extra, sinfo=sinfo)

    def set_logger(self):
        """设置logger输出格式以及日志输出路径"""
        if const.local_default_log_file_dir is None:
            const.get_local_default_log_file_dir()
        file_save_dir = os.path.expanduser(const.local_default_log_file_dir)

        validate_std_input_format(self.to_std, self.stdout_nodes, self.stdout_devices, self.stdout_level)
        validate_file_input_format(self.file_level, file_save_dir, self.append_rank_dir, self.file_name)

        rank_info = get_rank_info()

        self.to_std = judge_stdout(rank_info[0], rank_info[1], self.to_std, self.stdout_nodes, self.stdout_devices)
        if self.to_std:
            stream_handler = get_stream_handler(self.stdout_format, self.stdout_level)
            self.addHandler(stream_handler)

        file_path = get_file_path_list(file_save_dir, self.append_rank_dir, rank_info[2], rank_info[3], self.file_name)
        file_handlers = get_file_handler_list(self.file_level, file_path, self.max_file_size, self.max_num_of_files)
        for file_handler in file_handlers:
            self.addHandler(file_handler)
        self.setLevel(_convert_level('DEBUG'))
        self.set_flag = True

    # pylint: disable=W0221
    def _log(self,
             level,
             msg,
             args,
             **kwargs):
        """日志接口"""
        if not self.set_flag:
            try:
                self.set_logger()
            except (PathOwnerError, PathModeError, ValueError, MakeDirError, UnsupportedPlatformError, TypeError) as ex:
                if str(ex):
                    exc_type = sys.exc_info()
                    error_msg = f'An {str(exc_type[0])} occurred, {str(ex)}'
                    logging.error(error_msg)
                else:
                    logging.error(traceback.format_exc())
                sys.exit(ABNORMAL_EXIT_CODE)
            # pylint: disable=W0703
            except Exception:
                logging.error(traceback.format_exc())
                sys.exit(ABNORMAL_EXIT_CODE)
        super()._log(level, msg, args, **kwargs)


def set_logger_property(source):
    """设置日志实例的自定义属性"""
    for logger_name in logger_list:
        if source in LOGGER_SOURCE_LIST:
            logger_item = logger_list.get(logger_name)
            logger_item.source = source
            logger_item.pid = os.getpid()


def get_logger(logger_name: str = 'mx_logger', **kwargs) -> logging.Logger:
    """获取日志的实例

    Args:
        logger_name (str): 日志实例名称
        kwargs: 参数dict
            to_std (bool): 是否需要将日志内容输出到屏幕
            stdout_nodes (list[int] or tuple[int] or optional):
                需要输出日志内容到屏幕的节点列表
                默认: None, 表示所有节点的日志内容都会输出
                eg: [0, 1, 2, 3] or (0, 1, 2, 3): 表示0,1,2,3节点的日志内容都会输出到屏幕
            stdout_devices (list[int] or tuple[int] or optional):
                需要输出日志内容到屏幕的设备列表
                默认: None, 表示所有设备的日志内容都会输出
                eg: [0, 1, 2, 3] or (0, 1, 2, 3): 表示0,1,2,3设备的日志内容都会输出到屏幕
            stdout_level (str): 输出到屏幕的日志的等级，DEBUG, INFO, WARNING, ERROR, CRITICAL.
            stdout_format (str): 输出到屏幕的日志内容模板
            file_level (list[str] or tuple[str]): 落盘的日志的等级
                eg: ['INFO', 'ERROR'] 表示INFO以及ERROR等级以上的日志内容会落盘到对应的文件中
                列表的长度需要和文件名参数的长度保持一致
            append_rank_dir (bool): 是否需要按照rank进行日志区分
            file_name (list[str] or tuple[str]): 日志文件的名称列表
            max_file_size (int): 单个日志文件的大小阈值（MB）
            max_num_of_files (int): 日志绕接的文件数量的阈值

    Returns:
        mx_logger (logging.Logger): Logger
    """
    if logger_name in logger_list:
        return logger_list.get(logger_name)

    mx_logger = MxLogger(name=logger_name, **kwargs)
    logger_list[logger_name] = mx_logger

    return mx_logger


logger = get_logger('logger', to_std=True, file_name=['service.log'], file_level=['INFO'], append_rank_dir=True)
logger_without_std = get_logger('logger_without_std', to_std=False, file_name=['service.log'], file_level=['INFO'],
                                append_rank_dir=True)
operation_logger = get_logger('operation_logger', to_std=True, file_name=['operation.log'], file_level=['INFO'],
                              append_rank_dir=True)
operation_logger_without_std = get_logger('operation_logger_without_std', to_std=False, file_name=['operation.log'],
                                          file_level=['INFO'], append_rank_dir=True)


def record_operation_and_service_info_log(msg, *args, **kwargs):
    logger.info(msg, *args, **kwargs)
    operation_logger_without_std.info(msg, *args, **kwargs)


def record_operation_and_service_error_log(msg, *args, **kwargs):
    logger.error(msg, *args, **kwargs)
    operation_logger_without_std.error(msg, *args, **kwargs)


def record_operation_and_service_warning_log(msg, *args, **kwargs):
    logger.warning(msg, *args, **kwargs)
    operation_logger_without_std.warning(msg, *args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""MindPet SDK Module."""

from mindpet.mindpet_main import cli
from mindpet.log.log import logger, set_logger_property
from mindpet.utils.task_utils import handle_exception_log
from mindpet.utils.entrance_monitor import entrance_monitor
from mindpet.utils.constants import ENTRANCE_TYPE, EMPTY_STRING, ARG_NAMES, MINDPET_SDK_INTERFACE_NAMES

entrance_monitor.set_value(ENTRANCE_TYPE, 'SDK')
set_logger_property('SDK')


def finetune(*args, **kwargs):
    """
    SDK侧微调功能入口

    :param args: 参数args
    :param kwargs: 参数kwargs
    :return: 任务执行结果
    """
    return start_by_task_type(args, kwargs, task_type='finetune', ret_err_msg=False)


def evaluate(*args, **kwargs):
    """
    SDK侧评估功能入口

    :param args: 参数args
    :param kwargs: 参数kwargs
    :return: 任务执行结果
    """
    return start_by_task_type(args, kwargs, task_type='evaluate', ret_err_msg=EMPTY_STRING)


def infer(*args, **kwargs):
    """
    SDK侧推理功能入口

    :param args: 参数args
    :param kwargs: 参数kwargs
    :return: 任务执行结果
    """
    return start_by_task_type(args, kwargs, task_type='infer', ret_err_msg=EMPTY_STRING)


def start_by_task_type(args, kwargs, task_type, ret_err_msg):
    """
    根据任务类型启动任务

    :param task_type: 任务类型(finetune/evaluate/infer)
    :param ret_err_msg: 接口失败返回信息
    :param args: 参数args
    :param kwargs: 参数kwargs
    :return: 任务执行结果
    """
    if task_type not in MINDPET_SDK_INTERFACE_NAMES:
        logger.error('Invalid task_type for starting task.')
        return ret_err_msg

    try:
        commands = commands_generator(task_type, args, kwargs)
        return cli.main(commands, standalone_mode=False)
    # pylint: disable=W0719
    # pylint: disable=W0703
    except Exception as ex:
        handle_exception_log(ex)
        return ret_err_msg


def commands_generator(header, args, kwargs):
    """
    命令行生成器, 组装命令行参数

    :param header: 待执行功能
    :param args: 待组装args参数
    :param kwargs: 待组装kwargs参数
    :return: 组装后命令
    """
    output = []

    # 避免恶意传入过多参数导致下标越界
    args_length = min(len(args), len(ARG_NAMES.get(header)))

    for idx in range(args_length):
        output.append(f"--{(ARG_NAMES.get(header)[idx])}")
        output.append(f"--{(args[idx])}")

    for key_item, val_item in kwargs.items():
        # 安静模式仅允许CLI场景使用, SDK场景不允许配置该值, 给予错误警告
        if str(key_item) == 'quiet':
            raise ValueError('Param [quiet] is not supported by SDK.')
        # SDK场景参数值传None, 等价于CLI场景未传参数, 不应拼接到命令中
        if val_item is None:
            continue

        output.append(f"--{key_item}")
        output.append(f"--{val_item}")

    return [header] + output
